
/* 
   Yahoo Pager Client Library

   This code is based on code by Douglas Winslow. The original info from
   his code is listed below. This code has taken his code and has been 
   altered to my naming and coding conventions and has been made more 
   usable as a library of routines.

   -- Nathan Neulinger <nneul@umr.edu>
 */

/*
   Yahoo Pager Client Emulator Pro - yppro.c
   A basic reference implementation
   Douglas Winslow <douglas@min.net>
   Tue Sep  1 02:28:21 EDT 1998
   Version 2, Revision 2
   Known to compile on Linux 2.0, FreeBSD 2.2, and BSDi 3.0.
   hi to aap bdc drw jfn jrc mm mcd [cejn]b #cz and rootshell

   Finally!
   Yahoo finally patched their server-side, and things will be getting
   back to "normal".  I will continue to maintain this code as long as
   there is interest for it.  Since Yahoo will be discontinuing YPNS1.1
   login support shortly, I've upgraded this client to do YPNS1.2.  You
   *must* have a password to pass authentication to the pager server.
   This authentication is done by a weird HTTP cookie method.

   This code is distributed under the GNU General Public License (GPL)
 */

#include <stdio.h>
#include <netdb.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include "yahoolib.h"
#ifdef HAVE_DMALLOC
#include "dmalloc.h"
#else
#include <stdlib.h>
#endif

#if defined(WITH_SOCKS4)
void SOCKSinit(char *argv0);

#endif

/* pager server host */
#define YAHOO_PAGER_HOST	"jpager.yahoo.com"
#define YAHOO_PAGER_PORT	5050
/* pager server host for http connections */
#define YAHOO_PAGER_HTTP_HOST "http.pager.yahoo.com"
#define YAHOO_PAGER_HTTP_PORT 80
/* authentication/login host */
#define YAHOO_AUTH_HOST		"jpager.yahoo.com"
#define YAHOO_AUTH_PORT		80
/* buddy/identity/config host */
#define YAHOO_DATA_HOST		"jpager.yahoo.com"
#define YAHOO_DATA_PORT     80

/* User agent to use for HTTP connections */
/* It needs to have Mozilla/4 in it, otherwise it fails */
#ifndef VERSION
#define VERSION "1.0"
#endif
#define YAHOO_USER_AGENT	"YahooLib/" VERSION " (Mozilla/4.06)"

#define YAHOO_PROTOCOL_HEADER "YPNS1.4"

/* Service code labels for debugging output */
struct yahoo_idlabel yahoo_service_codes[] =
{
	{YAHOO_SERVICE_LOGON, "Pager Logon"},
	{YAHOO_SERVICE_LOGOFF, "Pager Logoff"},
	{YAHOO_SERVICE_ISAWAY, "Is Away"},
	{YAHOO_SERVICE_ISBACK, "Is Back"},
	{YAHOO_SERVICE_IDLE, "Idle"},
	{YAHOO_SERVICE_MESSAGE, "Message"},
	{YAHOO_SERVICE_IDACT, "Activate Identity"},
	{YAHOO_SERVICE_IDDEACT, "Deactivate Identity"},
	{YAHOO_SERVICE_MAILSTAT, "Mail Status"},
	{YAHOO_SERVICE_USERSTAT, "User Status"},
	{YAHOO_SERVICE_NEWMAIL, "New Mail"},
	{YAHOO_SERVICE_CHATINVITE, "Chat Invitation"},
	{YAHOO_SERVICE_CALENDAR, "Calendar Reminder"},
	{YAHOO_SERVICE_NEWPERSONALMAIL, "New Personals Mail"},
	{YAHOO_SERVICE_NEWCONTACT, "New Friend"},
	{YAHOO_SERVICE_GROUPRENAME, "Group Renamed"},
	{YAHOO_SERVICE_ADDIDENT, "Add Identity"},
	{YAHOO_SERVICE_ADDIGNORE, "Add Ignore"},
	{YAHOO_SERVICE_PING, "Ping"},
	{YAHOO_SERVICE_SYSMESSAGE, "System Message"},
	{YAHOO_SERVICE_CONFINVITE, "Conference Invitation"},
	{YAHOO_SERVICE_CONFLOGON, "Conference Logon"},
	{YAHOO_SERVICE_CONFDECLINE, "Conference Decline"},
	{YAHOO_SERVICE_CONFLOGOFF, "Conference Logoff"},
	{YAHOO_SERVICE_CONFMSG, "Conference Message"},
	{YAHOO_SERVICE_CHATLOGON, "Chat Logon"},
	{YAHOO_SERVICE_CHATLOGOFF, "Chat Logoff"},
	{YAHOO_SERVICE_CHATMSG, "Chat Message"},
	{0, NULL}
};

/* Status codes */
struct yahoo_idlabel yahoo_status_codes[] =
{
	{YAHOO_STATUS_AVAILABLE, "I'm Available"},
	{YAHOO_STATUS_BRB, "Be Right Back"},
	{YAHOO_STATUS_BUSY, "Busy"},
	{YAHOO_STATUS_NOTATHOME, "Not at Home"},
	{YAHOO_STATUS_NOTATDESK, "Not at my Desk"},
	{YAHOO_STATUS_NOTINOFFICE, "Not in the Office"},
	{YAHOO_STATUS_ONPHONE, "On the Phone"},
	{YAHOO_STATUS_ONVACATION, "On Vacation"},
	{YAHOO_STATUS_OUTTOLUNCH, "Out to Lunch"},
	{YAHOO_STATUS_STEPPEDOUT, "Stepped Out"},
	{YAHOO_STATUS_IDLE, "Idle"},
	{0, NULL}
};

/* Take a string and return the same string with all letters
   converted to lowercase */
char *strlower(char *in)
{
	char *tmp;
	int len, i;

	if (!in)
	{
		return NULL;
	}

	len = strlen(in);
	tmp = (char *) malloc(len + 1);

	for (i = 0; i < len; i++)
	{
		tmp[i] = tolower(in[i]);
	}
	tmp[i] = 0;

	return tmp;
}

/* Take a 4-byte character string in little-endian format and return
   a unsigned integer */
unsigned int yahoo_makeint(unsigned char *data)
{
	if (data)
	{
		return ((data[3] << 24) + (data[2] << 16) + (data[1] << 8) + (data[0]));
	}
	return 0;
}

/* Take an integer and store it into a 4 character little-endian string */
void yahoo_storeint(unsigned char *data, unsigned int val)
{
	unsigned int tmp = val;
	int i;

	if (data)
	{
		for (i = 0; i < 4; i++)
		{
			data[i] = tmp % 256;
			tmp >>= 8;
		}
	}
}

/* Initialize interface to yahoo library, sortof like a class object
   creation routine. */
struct yahoo_context *yahoo_init(char *user, char *password, struct yahoo_options *options)
{
	struct yahoo_context *tmp;

	if (!user || !password)
	{
		return NULL;
	}

	/* Allocate a new context */
	tmp = (struct yahoo_context *) calloc(1, sizeof(*tmp));

	/* Fill in any available info */
	tmp->user = strdup(user);
	tmp->password = strdup(password);
	if (options->proxy_host)
	{
		tmp->proxy_host = strdup(options->proxy_host);
	}
	tmp->proxy_port = options->proxy_port;
	tmp->connect_mode = options->connect_mode;

#if defined(WITH_SOCKS4)
	if (connect_mode == YAHOO_CONNECT_SOCKS4)
	{
		static int did_socks_init = 0;

		if (did_socks_init == 0)
		{
			SOCKSinit("YahooLib");
			did_socks_init = 1;
		}
	}
#endif

	if (options->debug > 0)
	{
		printf("YahooLib: debugging enabled at level (%d)\n", options->debug);
		tmp->debug = options->debug;
	}

	/* Fetch the cookies */
	yahoo_fetchcookies(tmp);

	return tmp;
}

/* Turn a status code into it's corresponding string */
char *yahoo_get_status_string(int statuscode)
{
	int i;

	for (i = 0; yahoo_status_codes[i].label; i++)
	{
		if (yahoo_status_codes[i].id == statuscode)
		{
			return yahoo_status_codes[i].label;
		}
	}
	return NULL;
}

/* Turn a service code into it's corresponding string */
char *yahoo_get_service_string(int servicecode)
{
	int i;
	char *name = "Unknown Service";
	static char tmp[50];

	for (i = 0; yahoo_service_codes[i].label; i++)
	{
		if (yahoo_service_codes[i].id == servicecode)
		{
			name = yahoo_service_codes[i].label;
			break;
		}
	}

	sprintf(tmp, "(%d) %s", servicecode, name);
	return tmp;
}

/* Return a malloc()'d copy of the users cookie */
int yahoo_fetchcookies(struct yahoo_context *ctx)
{
	char buffer[5000], content_len[32];
	int servfd;
	int i;
	char *tmpstr;

	if (!ctx)
	{
		return 0;
	}

	if (ctx->debug > 0)
	{
		printf("YahooLib: Fetching login cookies\n");
	}

	/* Check for cached cookie */
	if (ctx->cookie)
	{
		free(ctx->cookie);
		ctx->cookie = 0;
	}
	if (ctx->login_cookie)
	{
		free(ctx->login_cookie);
		ctx->login_cookie = 0;
	}

	if (ctx->connect_mode == YAHOO_CONNECT_HTTPPROXY)
	{
		servfd = yahoo_socket_connect(ctx, ctx->proxy_host,
			ctx->proxy_port);
	}
	else
	{
		servfd = yahoo_socket_connect(ctx, YAHOO_AUTH_HOST,
			YAHOO_AUTH_PORT);
	}
	if (!servfd)
	{
		printf("YahooLib: failed to connect to pager auth server.\n");
		return (0);
	}

	strcpy(buffer, "POST ");
	if (ctx->connect_mode == YAHOO_CONNECT_HTTPPROXY)
	{
		strcat(buffer, "http://" YAHOO_AUTH_HOST);
	}
	strcat(buffer, "/config/login HTTP/1.0\r\n");

	if (ctx->login_id)
	{
		sprintf(content_len, "Content-length: %i\r\n\r\n",
			strlen(yahoo_urlencode(ctx->login_id)) +
			strlen(yahoo_urlencode(ctx->password)) + 14);
	}
	else
	{
		sprintf(content_len, "Content-length: %i\r\n\r\n",
			strlen(yahoo_urlencode(ctx->user)) +
			strlen(yahoo_urlencode(ctx->password)) + 14);
	}

	/* The stupid server won't accept the request unless it detects
	   'Mozilla/4' in the User-Agent - would be really nice to be able
	   to generate the cookie ourself. */
	strcat(buffer, "User-Agent: YahooLib/1.0 (Mozilla/4.06)\r\n");

	strcat(buffer, "Host: edit.my.yahoo.com\r\n");
	strcat(buffer, "Content-type: application/x-www-form-urlencoded\r\n");
	strcat(buffer, content_len);
	strcat(buffer, "login=");
	if (ctx->login_id)
	{
		strcat(buffer, yahoo_urlencode(ctx->login_id));
	}
	else
	{
		strcat(buffer, yahoo_urlencode(ctx->user));
	}
	strcat(buffer, "&passwd=");
	strcat(buffer, yahoo_urlencode(ctx->password));
	strcat(buffer, "\r\n");

	write(servfd, buffer, strlen(buffer));

	if (ctx->debug > 2)
	{
		printf("YahooLib: fetchcookies - writing buffer '%s'\n", buffer);
	}

	ctx->cookie = NULL;
	while (yahoo_tcp_readline(buffer, 5000, servfd) > 0)
	{
		/* strip out any non-alphas */
		for (i = 0; i < strlen(buffer); i++)
		{
			if (!isprint(buffer[i]))
			{
				buffer[i] = 0;
			}
		}

		if (ctx->debug > 2)
			if (ctx->debug > 2)
			{
				printf("YahooLib: fetchcookies read buffer '%s'\n", buffer);
			}

		if (!strncmp(buffer, "Set-Cookie: Y=", 14))
		{
			if (ctx->cookie)
			{
				free(ctx->cookie);
			}
			ctx->cookie = strdup(buffer + 12);

			tmpstr = strchr(ctx->cookie, ';');
			if (tmpstr)
			{
				*tmpstr = '\0';
			}
		}
	}
	close(servfd);

	if (ctx->cookie)
	{
		tmpstr = strstr(ctx->cookie, "n=");
		if (tmpstr)
		{
			ctx->login_cookie = strdup(tmpstr + 2);
		}

		tmpstr = strchr(ctx->login_cookie, '&');
		if (tmpstr)
		{
			*tmpstr = '\0';
		}
	}

	if (ctx->debug > 1)
	{
		if (ctx->cookie)
			printf("YahooLib: Cookie (%s)\n", ctx->cookie);
		if (ctx->login_cookie)
			printf("YahooLib: Login Cookie (%s)\n", ctx->login_cookie);
	}

	if (ctx->debug > 0)
	{
		printf("YahooLib: Done fetching login cookies\n");
	}

	return 0;
}

/* Add a buddy to your buddy list */
int yahoo_add_buddy(struct yahoo_context *ctx, char *addid,
	char *active_id, char *group, char *msg)
{
	char buffer[5000];
	int servfd;

	if (!ctx)
	{
		return 0;
	}

	if (ctx->connect_mode == YAHOO_CONNECT_HTTPPROXY)
	{
		servfd = yahoo_socket_connect(ctx, ctx->proxy_host,
			ctx->proxy_port);
	}
	else
	{
		servfd = yahoo_socket_connect(ctx, YAHOO_DATA_HOST,
			YAHOO_DATA_PORT);
	}
	if (!servfd)
	{
		printf("YahooLib: failed to connect to pager data server.\n");
		return (0);
	}

	strcpy(buffer, "GET ");
	if (ctx->connect_mode == YAHOO_CONNECT_HTTPPROXY)
	{
		strcat(buffer, "http://" YAHOO_DATA_HOST);
	}
	strcat(buffer, "/config/set_buddygrp?.bg=");
	strcat(buffer, yahoo_urlencode(group));
	strcat(buffer, "&.src=bl&.cmd=a&.bdl=");
	strcat(buffer, yahoo_urlencode(addid));
	strcat(buffer, "&.id=");
	strcat(buffer, yahoo_urlencode(active_id));
	strcat(buffer, "&.l=");
	strcat(buffer, yahoo_urlencode(ctx->user));
	strcat(buffer, "&.amsg=");
	strcat(buffer, yahoo_urlencode(msg));
	strcat(buffer, " HTTP/1.0\r\n");

	strcat(buffer, "User-Agent: " YAHOO_USER_AGENT "\r\n");
	strcat(buffer, "Host: jpager.yahoo.com\r\n");
	strcat(buffer, "Cookie: ");
	strcat(buffer, ctx->cookie);
	strcat(buffer, "\r\n");
	strcat(buffer, "\r\n");

	write(servfd, buffer, strlen(buffer));
	while (yahoo_tcp_readline(buffer, 5000, servfd) > 0)
	{
		/* just dump the output, I don't care about errors at the moment */
	}
	close(servfd);

	/* indicate success for now with 0 */
	return 0;
}

/* Remove a buddy from your buddy list */
int yahoo_remove_buddy(struct yahoo_context *ctx, char *addid,
	char *active_id, char *group, char *msg)
{
	char buffer[5000];
	int servfd;

	if (!ctx)
	{
		return 0;
	}

	servfd = yahoo_socket_connect(ctx, YAHOO_DATA_HOST, YAHOO_DATA_PORT);
	if (!servfd)
	{
		printf("YahooLib: failed to connect to data server.\n");
		return (0);
	}

	strcpy(buffer, "GET /config/set_buddygrp?.bg=");
	strcat(buffer, yahoo_urlencode(group));
	strcat(buffer, "&.src=bl&.cmd=d&.bdl=");
	strcat(buffer, yahoo_urlencode(addid));
	strcat(buffer, "&.id=");
	strcat(buffer, yahoo_urlencode(active_id));
	strcat(buffer, "&.l=");
	strcat(buffer, yahoo_urlencode(ctx->user));
	strcat(buffer, "&.amsg=");
	strcat(buffer, yahoo_urlencode(msg));
	strcat(buffer, " HTTP/1.0\r\n");

	strcat(buffer, "User-Agent: " YAHOO_USER_AGENT "\r\n");
	strcat(buffer, "Host: jpager.yahoo.com\r\n");
	strcat(buffer, "Cookie: ");
	strcat(buffer, ctx->cookie);
	strcat(buffer, "\r\n");
	strcat(buffer, "\r\n");

	write(servfd, buffer, strlen(buffer));
	while (yahoo_tcp_readline(buffer, 5000, servfd) > 0)
	{
		/* just dump the output, I don't care about errors at the moment */
	}
	close(servfd);

	/* indicate success for now with 0 */
	return 0;
}

/* Retrieve the configuration from the server */
int yahoo_get_config(struct yahoo_context *ctx)
{
	char buffer[5000];
	int i;
	int servfd;
	int commas;
	int in_section;
	struct yahoo_buddy **buddylist = NULL;
	int buddycnt = 0;
	int nextbuddy = 0;

	/* Check for cached cookie */
	if (!ctx || !ctx->cookie)
	{
		return 0;
	}

	if (ctx->debug > 0)
	{
		printf("YahooLib: Fetching config.\n");
	}

	if (ctx->connect_mode == YAHOO_CONNECT_HTTPPROXY)
	{
		servfd = yahoo_socket_connect(ctx, ctx->proxy_host,
			ctx->proxy_port);
	}
	else
	{
		servfd = yahoo_socket_connect(ctx, YAHOO_DATA_HOST,
			YAHOO_DATA_PORT);
	}
	if (!servfd)
	{
		printf("YahooLib: failed to connect to pager data server.\n");
		return (0);
	}

	strcpy(buffer, "GET ");
	if (ctx->connect_mode == YAHOO_CONNECT_HTTPPROXY)
	{
		strcat(buffer, "http://" YAHOO_DATA_HOST);
	}
	strcat(buffer, "/config/get_buddylist?.src=bl HTTP/1.0\r\n");
	strcat(buffer, "Cookie: ");
	strcat(buffer, ctx->cookie);
	strcat(buffer, "\r\n");
	strcat(buffer, "\r\n");

	write(servfd, buffer, strlen(buffer));
	if (ctx->debug > 2)
	{
		printf("YahooLib: getconfig sending buffer '%s'\n", buffer);
	}

	in_section = 0;
	while (yahoo_tcp_readline(buffer, 5000, servfd) > 0)
	{
		/* strip out any non-alphas */
		for (i = 0; i < strlen(buffer); i++)
		{
			if (!isprint(buffer[i]))
			{
				buffer[i] = 0;
			}
		}

		if (ctx->debug > 2)
		{
			printf("YahooLib: getconfig read: %s\n", buffer);
		}

		if (!strcasecmp(buffer, "BEGIN IDENTITIES"))
		{
			in_section = 1;
		}
		else if (!strcasecmp(buffer, "END IDENTITIES"))
		{
			in_section = 0;
		}
		else if (!strcasecmp(buffer, "BEGIN BUDDYLIST"))
		{
			in_section = 2;
		}
		else if (!strcasecmp(buffer, "END BUDDYLIST"))
		{
			in_section = 0;
		}
		else if (in_section == 1)
		{
			char *tmp;

			/* count the commas */
			commas = 0;
			for (i = 0; i < strlen(buffer); i++)
			{
				if (buffer[i] == ',')
				{
					commas++;
				}
			}

			/* make sure we've gotten rid of any previous identities array */
			yahoo_free_identities(ctx);

			/* allocate the array to hold the list of identities */
			ctx->identities = (char **) calloc(commas + 2, sizeof(char *));

			/* Parse through the list and get all the entries */
			i = 0;
			tmp = strtok(buffer, ",");
			while (tmp)
			{
				if (ctx->debug > 0)
				{
					printf("YahooLib: Retrieved identity '%s'\n", tmp);
				}
				ctx->identities[i++] = strdup(tmp);
				tmp = strtok(NULL, ",");
			}
			ctx->identities[i] = 0;
		}
		else if (in_section == 2)
		{
			char *group;
			char *tmp;
			struct yahoo_buddy **tmp_buddylist;
			struct yahoo_buddy *tmpbuddy;
			int tmp_buddycnt;

			/* count the buddies on this line */
			tmp_buddycnt = buddycnt;
			for (i = 0; i < strlen(buffer); i++)
			{
				if (buffer[i] == ',')
				{
					buddycnt++;
				}
			}
			buddycnt++;			/* always one more than comma count */

			/* allocate the array to hold the list of buddy */
			tmp_buddylist = (struct yahoo_buddy **)
				malloc(sizeof(struct yahoo_buddy *) * (buddycnt + 1));

			/* Free and copy the old one if necessary */
			if (buddylist)
			{
				memcpy(tmp_buddylist, buddylist,
					(tmp_buddycnt + 1) * sizeof(struct yahoo_buddy *));

				free(buddylist);
			}
			buddylist = tmp_buddylist;

			/* Parse through the list and get all the entries */
			tmp = strtok(buffer, ",:");
			group = NULL;
			while (tmp)
			{
				if (tmp == buffer)	/* group name */
				{
					group = tmp;
				}
				else
				{
					tmpbuddy = (struct yahoo_buddy *)
						malloc(sizeof(struct yahoo_buddy));

					tmpbuddy->id = strdup(tmp);
					tmpbuddy->group = strdup(group);
					if (ctx->debug > 0)
					{
						printf("YahooLib: Retrieved buddy '%s:%s'\n",
							group, tmp);
					}
					buddylist[nextbuddy++] = tmpbuddy;
				}
				tmp = strtok(NULL, ",");
			}
			buddylist[nextbuddy] = 0;
		}
		else if (!strncasecmp(buffer, "Mail=", strlen("Mail=")))
		{
			ctx->mail = atoi(buffer + strlen("Mail="));
			if (ctx->debug > 0)
			{
				printf("YahooLib: Retrieve mail flag '%d'\n", ctx->mail);
			}
		}
		else if (!strncasecmp(buffer, "Login=", strlen("Login=")))
		{
			if (ctx->login_id)
			{
				free(ctx->login_id);
			}
			ctx->login_id = strdup(buffer + strlen("Login="));
			if (ctx->debug > 0)
			{
				printf("YahooLib: Retrieve login id '%s'\n", ctx->login_id);
			}
		}
	}
	close(servfd);

	yahoo_free_buddies(ctx);
	ctx->buddies = buddylist;

	/* fill in a bogus login_in, just in case */
	if (!ctx->login_id)
	{
		ctx->login_id = strdup(ctx->user);
	}

	if (ctx->debug > 0)
	{
		printf("YahooLib: Done fetching config.\n");
	}

	/* refetch the cookie if the login_id is different so that
	   it will have the correct info in it */
	if (strcmp(ctx->login_id, ctx->user))
	{
		yahoo_fetchcookies(ctx);
	}

	return 1;
}

/* Free a buddy list */
void yahoo_free_buddies(struct yahoo_context *ctx)
{
	int i;

	if (!ctx->buddies)
	{
		return;
	}

	i = 0;
	while (ctx->buddies[i])
	{
		if (ctx->buddies[i]->group)
			free(ctx->buddies[i]->group);
		if (ctx->buddies[i]->id)
			free(ctx->buddies[i]->id);
		i++;
	}

	free(ctx->buddies);
	ctx->buddies = NULL;
}

/* Free a identities list */
void yahoo_free_identities(struct yahoo_context *ctx)
{
	int i;

	if (ctx->debug > 1)
	{
		printf("YahooLib: Freeing identity list.\n");
	}

	if (!ctx->identities)
	{
		return;
	}

	i = 0;
	while (ctx->identities[i])
	{
		if (ctx->identities[i])
			free(ctx->identities[i]);
		ctx->identities[i] = NULL;
		i++;
	}

	free(ctx->identities);
	ctx->identities = NULL;
}

/* Log in, optionally activating other secondary identities */
int yahoo_cmd_logon(struct yahoo_context *ctx)
{
	char login_string[5000];	/* need to change to malloc ASAP */
	char *tmpid;
	char **identities = ctx->identities;
	int i;

	if (!ctx || !ctx->login_cookie)
	{
		printf("yahoo_cmd_logon: logon called without context and/or cookie.\n");
		exit(1);
	}

	strcpy(login_string, ctx->login_cookie);
/* testing with new logon code */
//  strcpy(login_string, "$1$_2S43d5f$XXXXXXXXWtRKNclLWyy8C.");

	login_string[strlen(login_string) + 1] = 0;
	login_string[strlen(login_string)] = 1;		/* control-A */

	strcat(login_string, ctx->user);

	/* Send all identities */
	if (identities)
	{
		i = 0;
		tmpid = identities[i];
		while (tmpid)
		{
			if (strcasecmp(tmpid, ctx->user))
			{
				strcat(login_string, ",");
				strcat(login_string, tmpid);
			}
			tmpid = identities[i++];
		}
	}

	yahoo_sendcmd(ctx, YAHOO_SERVICE_LOGON, ctx->user, login_string);

	/* something that the windows one sends, not sure what it is */
	login_string[0] = 0;
	strcat(login_string, "C=0\002");
	strcat(login_string, "F=0,P=0,H=0,S=0,W=0,O=0\002");
	strcat(login_string, "M=0,P=0,C=0,S=0");
	yahoo_sendcmd(ctx, YAHOO_SERVICE_PASSTHROUGH2, ctx->user, login_string);

	return 0;
}

int yahoo_connect(struct yahoo_context *ctx)
{
	int res;

	res = 0;
	ctx->sockfd = 0;

	switch (ctx->connect_mode)
	{
		case YAHOO_CONNECT_SOCKS4:
		case YAHOO_CONNECT_SOCKS5:
		case YAHOO_CONNECT_NORMAL:
			ctx->sockfd = yahoo_socket_connect(ctx, YAHOO_PAGER_HOST,
				YAHOO_PAGER_PORT);
			if (!ctx->sockfd)
			{
				printf("YahooLib: couldn't connect to pager host\n");
				return (0);
			}
			break;

		case YAHOO_CONNECT_HTTP:
		case YAHOO_CONNECT_HTTPPROXY:
			/* no pager connection will be established for this */
			break;

		default:
			printf("YahooLib: unhandled connect mode (%d)\n",
				ctx->connect_mode);
	}

	return (1);
}

/* Send a packet to the server via http connection method */
/* at moment only handles regular http connection, once I have that 
   working, this code needs to also do http proxy connections as well */
int yahoo_sendcmd_http(struct yahoo_context *ctx, struct yahoo_rawpacket *pkt)
{
	int sockfd;
	char buffer[5000];
	char tmpbuf[1000];
	int size;
	int res;

	if (!ctx || !pkt)
	{
		return (0);
	}

	size = YAHOO_PACKET_HEADER_SIZE + strlen(pkt->content) + 1;

	if (ctx->connect_mode == YAHOO_CONNECT_HTTPPROXY)
	{
		sockfd = yahoo_socket_connect(ctx, ctx->proxy_host,
			ctx->proxy_port);
	}
	else
	{
		sockfd = yahoo_socket_connect(ctx, YAHOO_PAGER_HTTP_HOST,
			YAHOO_PAGER_HTTP_PORT);
	}
	if (!sockfd)
	{
		printf("YahooLib: failed to connect to pager http server.\n");
		return (0);
	}

	strcpy(buffer, "POST ");
	if (ctx->connect_mode == YAHOO_CONNECT_HTTPPROXY)
	{
		strcat(buffer, "http://" YAHOO_PAGER_HTTP_HOST);
	}
	strcat(buffer, "/notify HTTP/1.0\r\n");

	strcat(buffer, "Host: " YAHOO_PAGER_HTTP_HOST "\r\n");
//  strcat(buffer, "Referer: hostname\r\n");
	//  strcat(buffer, "Accept: application/octet-stream\r\n");

	sprintf(tmpbuf, "Content-Length: %d\r\n", size);
	strcat(buffer, tmpbuf);

	strcat(buffer, "User-Agent: " YAHOO_USER_AGENT "\r\n");
	strcat(buffer, "Pragma: No-Cache\r\n");

	strcat(buffer, "Cookie: ");
	strcat(buffer, ctx->cookie);
	strcat(buffer, "\r\n");
	strcat(buffer, "\r\n");

	write(sockfd, buffer, strlen(buffer));
	write(sockfd, pkt, size);
	write(sockfd, "\r\n", 2);

	/* now we need to read the results */
	/* I'm taking the cheat approach and just dumping them onto the
	   buffer, headers and all, the _skip_to_YHOO_ code will handle it 
	   for now */

	while ((res = read(sockfd, buffer, 5000)) > 0)
	{
		if (res == -1)
		{
			printf("YahooLib: Error reading data from server.\n");
			exit(1);
		}
		yahoo_addtobuffer(ctx, buffer, res);
	}
	close(sockfd);

	return (0);
}

/* Send a packet to the server, called by all routines that want to issue
   a command. */
int yahoo_sendcmd(struct yahoo_context *ctx, int service, char *active_nick, char *content)
{
	int size;
	struct yahoo_rawpacket *pkt;

	/* why the )&*@#$( did they hardwire the packet size that gets sent
	   when the size of the packet is included in what is sent, bizarre */
	size = 4 * 256 + YAHOO_PACKET_HEADER_SIZE;
	pkt = (struct yahoo_rawpacket *) calloc(1, size);

	/* Build the packet */
	strcpy(pkt->version, YAHOO_PROTOCOL_HEADER);
	yahoo_storeint(pkt->len, size);
	yahoo_storeint(pkt->service, service);

	/* Not sure, but might as well send for regular connections as well. */
	yahoo_storeint(pkt->magic_id, ctx->magic_id);
	// setting this causes us to get knocked off
	//yahoo_storeint(pkt->connection_id, ctx->connection_id);

	strcpy(pkt->nick1, ctx->login_id);
	strcpy(pkt->nick2, active_nick);
	strcpy(pkt->content, content);

	if (ctx->debug > 1)
	{
		yahoo_hexdump("send_cmd", (char *) pkt, size);
	}

	switch (ctx->connect_mode)
	{
		case YAHOO_CONNECT_SOCKS4:
		case YAHOO_CONNECT_SOCKS5:
		case YAHOO_CONNECT_NORMAL:
			write(ctx->sockfd, pkt, size);
			break;
		case YAHOO_CONNECT_HTTP:
		case YAHOO_CONNECT_HTTPPROXY:
			yahoo_sendcmd_http(ctx, pkt);
			break;
	}

	free(pkt);
	return (0);
}

int yahoo_cmd_ping(struct yahoo_context *ctx)
{
	yahoo_sendcmd(ctx, YAHOO_SERVICE_PING, ctx->user, "");
	return (0);
}

int yahoo_cmd_idle(struct yahoo_context *ctx)
{
	yahoo_sendcmd(ctx, YAHOO_SERVICE_IDLE, ctx->user, "");
	return (0);
}

int yahoo_cmd_msg(struct yahoo_context *ctx, char *active_user,
	char *touser, char *msg)
{
	char *content;

	content = (char *) malloc(strlen(touser) + strlen(msg) + 5);

	if (strlen(touser))
	{
		sprintf(content, "%s,%s", touser, msg);
		yahoo_sendcmd(ctx, YAHOO_SERVICE_MESSAGE, active_user, content);
	}

	free(content);
	return (0);
}

int yahoo_cmd_set_away_mode(struct yahoo_context *ctx, int status)
{
	char statusstring[50];

	if (ctx->debug)
	{
		printf("YahooLib: sending ISAWAY for (%s) with status (%d)\n",
			ctx->user, status);
	}

	sprintf(statusstring, "%d", status);
	yahoo_sendcmd(ctx, YAHOO_SERVICE_ISAWAY, ctx->user, statusstring);

	return 0;
}

int yahoo_cmd_set_back_mode(struct yahoo_context *ctx, int status)
{
	char statusstring[50];

	if (ctx->debug)
	{
		printf("YahooLib: sending ISBACK for (%s) with status (%d)\n",
			ctx->user, status);
	}

	sprintf(statusstring, "%d", status);
	yahoo_sendcmd(ctx, YAHOO_SERVICE_ISBACK, ctx->user, statusstring);

	return 0;
}

int yahoo_cmd_activate_id(struct yahoo_context *ctx, char *newid)
{
	if (strlen(newid))
	{
		yahoo_sendcmd(ctx, YAHOO_SERVICE_IDACT, newid, newid);
	}
	return (0);
}

int yahoo_cmd_user_status(struct yahoo_context *ctx)
{
	yahoo_sendcmd(ctx, YAHOO_SERVICE_USERSTAT, ctx->user, "");
	return (0);
}

int yahoo_cmd_logoff(struct yahoo_context *ctx)
{
	yahoo_sendcmd(ctx, YAHOO_SERVICE_LOGOFF, ctx->user, ctx->user);
	return (0);
}

/*
 * Free the rawpacket structure - primarily a placeholder
 * since all static elements at the moment
 */
void yahoo_free_rawpacket(struct yahoo_rawpacket *pkt)
{
	if (pkt)
	{
		free(pkt);
	}
}

/*
 * Free entire packet structure including string elements
 */
void yahoo_free_packet(struct yahoo_packet *pkt)
{
	int i;

	if (pkt)
	{
		if (pkt->real_id)
		{
			free(pkt->real_id);
		}
		if (pkt->active_id)
		{
			free(pkt->active_id);
		}
		if (pkt->idstatus)
		{
			for (i = 0; i < pkt->idstatus_count; i++)
			{
				yahoo_free_idstatus(pkt->idstatus[i]);
			}
			free(pkt->idstatus);
		}
		if (pkt->msg)
		{
			free(pkt->msg);
		}
		free(pkt);
	}
}

void yahoo_free_idstatus(struct yahoo_idstatus *idstatus)
{
	if (idstatus->id)
	{
		free(idstatus->id);
	}
	if (idstatus->connection_id)
	{
		free(idstatus->connection_id);
	}
	free(idstatus);
}

struct yahoo_packet *yahoo_parsepacket(struct yahoo_context *ctx, struct yahoo_rawpacket *inpkt)
{
	struct yahoo_packet *pkt;

	/* If no valid inpkt passed, return */
	if (!inpkt)
	{
		return NULL;
	}

	/* Allocate the packet structure, zeroed out */
	pkt = (struct yahoo_packet *) calloc(sizeof(*pkt), 1);

	/* Pull out the standard data */
	pkt->service = yahoo_makeint(inpkt->service);
	pkt->connection_id = yahoo_makeint(inpkt->connection_id);
	pkt->real_id = strdup(inpkt->nick1);
	pkt->active_id = strdup(inpkt->nick2);

	pkt->magic_id = yahoo_makeint(inpkt->magic_id);
	pkt->unknown1 = yahoo_makeint(inpkt->unknown1);
	pkt->unknown2 = yahoo_makeint(inpkt->unknown2);

	/* doing this seems like a cleaner approach, but am not sure if it is
	   a valid one */
	if (pkt->magic_id != 0)
	{
		ctx->magic_id = pkt->magic_id;
	}
	if (pkt->connection_id != 0)
	{
		ctx->connection_id = pkt->connection_id;
	}

	/* Call a particular parse routine to pull out the content */
	switch (pkt->service)
	{
		case YAHOO_SERVICE_LOGON:
		case YAHOO_SERVICE_LOGOFF:
		case YAHOO_SERVICE_ISAWAY:
		case YAHOO_SERVICE_ISBACK:
		case YAHOO_SERVICE_USERSTAT:
		case YAHOO_SERVICE_CHATLOGON:
		case YAHOO_SERVICE_CHATLOGOFF:
			yahoo_parsepacket_status(ctx, pkt, inpkt);
			break;
		case YAHOO_SERVICE_IDACT:
		case YAHOO_SERVICE_IDDEACT:
			/* nothing needs done, only has main fields */
			break;
		case YAHOO_SERVICE_MESSAGE:
		case YAHOO_SERVICE_SYSMESSAGE:
		case YAHOO_SERVICE_CHATMSG:
			yahoo_parsepacket_message(ctx, pkt, inpkt);
			break;
		case YAHOO_SERVICE_NEWMAIL:
		case YAHOO_SERVICE_NEWPERSONALMAIL:
			yahoo_parsepacket_newmail(ctx, pkt, inpkt);
			break;
		case YAHOO_SERVICE_CALENDAR:
			yahoo_parsepacket_calendar(ctx, pkt, inpkt);
			break;
		case YAHOO_SERVICE_CHATINVITE:
			yahoo_parsepacket_chatinvite(ctx, pkt, inpkt);
			break;
		case YAHOO_SERVICE_NEWCONTACT:
			yahoo_parsepacket_newcontact(ctx, pkt, inpkt);
			break;
		case YAHOO_SERVICE_GROUPRENAME:
			yahoo_parsepacket_grouprename(ctx, pkt, inpkt);
			break;
		case YAHOO_SERVICE_CONFINVITE:
			yahoo_parsepacket_conference_invite(ctx, pkt, inpkt);
			break;
		case YAHOO_SERVICE_CONFLOGON:
		case YAHOO_SERVICE_CONFLOGOFF:
			yahoo_parsepacket_conference_user(ctx, pkt, inpkt);
			break;
		case YAHOO_SERVICE_CONFMSG:
			yahoo_parsepacket_conference_msg(ctx, pkt, inpkt);
			break;
		case YAHOO_SERVICE_PING:
			yahoo_parsepacket_ping(ctx, pkt, inpkt);
			break;
		default:
			printf("YahooLib: can't parse packet type (%d)\n", pkt->service);
			break;
	}

	return pkt;
}

int yahoo_parsepacket_ping(struct yahoo_context *ctx, struct yahoo_packet *pkt, struct yahoo_rawpacket *inpkt)
{
	char *content;

	/* Make working copy of content */
	content = inpkt->content;

	pkt->msg = NULL;
	if (content)
	{
		pkt->msg = strdup(content);
	}

	return 0;
}

int yahoo_parsepacket_newmail(struct yahoo_context *ctx, struct yahoo_packet *pkt, struct yahoo_rawpacket *inpkt)
{
	char *content;
	int len;

	/* Make working copy of content */
	content = inpkt->content;
	len = strlen(content);

	if (pkt->service == YAHOO_SERVICE_NEWMAIL)
	{
		pkt->mail_status = 0;
		if (len > 0)
		{
			pkt->mail_status = atoi(content);
		}
	}
	else if (pkt->service == YAHOO_SERVICE_NEWPERSONALMAIL)
	{
		pkt->mail_status = 0;
		if (len > 0)
		{
			pkt->mail_status = atoi(content);
		}
	}

	return 0;
}

int yahoo_parsepacket_grouprename(struct yahoo_context *ctx, struct yahoo_packet *pkt, struct yahoo_rawpacket *inpkt)
{
	char *content;
	char *tmp, delim[5];

	/* Make working copy of content */
	content = strdup(inpkt->content);

	/* init elements to all null */
	pkt->group_old = NULL;
	pkt->group_new = NULL;

	tmp = NULL;
	delim[0] = 1;				/* control-a */
	delim[1] = 0;

	if (content)
	{
		tmp = strtok(content, delim);
	}

	if (tmp)					/* got the conference id */
	{
		pkt->group_old = strdup(tmp);
		tmp = strtok(NULL, delim);
	}

	if (tmp)					/* conference host */
	{
		pkt->group_new = strdup(tmp);
		tmp = strtok(NULL, delim);
	}

	free(content);
	return (0);
}

int yahoo_parsepacket_conference_invite(struct yahoo_context *ctx, struct yahoo_packet *pkt, struct yahoo_rawpacket *inpkt)
{
	char *content;
	char *tmp, delim[5];

	/* Make working copy of content */
	content = strdup(inpkt->content);

	/* init elements to all null */
	pkt->conf_id = NULL;
	pkt->conf_host = NULL;
	pkt->conf_user = NULL;
	pkt->conf_msg = NULL;

	tmp = NULL;
	delim[0] = 2;				/* control-b */
	delim[1] = 0;

	if (content)
	{
		tmp = strtok(content, delim);
	}

	if (tmp)					/* got the conference id */
	{
		pkt->conf_id = strdup(tmp);
		tmp = strtok(NULL, delim);
	}

	if (tmp)					/* conference host */
	{
		pkt->conf_id = strdup(tmp);
		tmp = strtok(NULL, delim);
	}

	if (tmp)					/* conference user */
	{
		pkt->conf_user = strdup(tmp);
		tmp = strtok(NULL, delim);
	}

	if (tmp)					/* msg */
	{
		pkt->conf_msg = strdup(tmp);
		tmp = strtok(NULL, delim);
	}

	free(content);
	return 0;
}

int yahoo_parsepacket_conference_msg(struct yahoo_context *ctx, struct yahoo_packet *pkt, struct yahoo_rawpacket *inpkt)
{
	char *content;
	char *tmp, delim[5];

	/* Make working copy of content */
	content = strdup(inpkt->content);

	/* init elements to all null */
	pkt->conf_id = NULL;
	pkt->conf_host = NULL;
	pkt->conf_user = NULL;
	pkt->conf_msg = NULL;

	tmp = NULL;
	delim[0] = 2;				/* control-b */
	delim[1] = 0;

	if (content)
	{
		tmp = strtok(content, delim);
	}

	if (tmp)					/* got the conference id */
	{
		pkt->conf_id = strdup(tmp);
		tmp = strtok(NULL, delim);
	}

	if (tmp)					/* conference user */
	{
		pkt->conf_user = strdup(tmp);
		tmp = strtok(NULL, delim);
	}

	if (tmp)					/* msg */
	{
		pkt->conf_msg = strdup(tmp);
		tmp = strtok(NULL, delim);
	}

	free(content);
	return 0;
}

int yahoo_parsepacket_conference_user(struct yahoo_context *ctx, struct yahoo_packet *pkt, struct yahoo_rawpacket *inpkt)
{
	char *content;
	char *tmp, delim[5];

	/* Make working copy of content */
	content = strdup(inpkt->content);

	/* init elements to all null */
	pkt->conf_id = NULL;
	pkt->conf_host = NULL;
	pkt->conf_user = NULL;
	pkt->conf_msg = NULL;

	tmp = NULL;
	delim[0] = 2;				/* control-b */
	delim[1] = 0;

	if (content)
	{
		tmp = strtok(content, delim);
	}

	if (tmp)					/* got the conference id */
	{
		pkt->conf_id = strdup(tmp);
		tmp = strtok(NULL, delim);
	}

	if (tmp)					/* conference user */
	{
		pkt->conf_user = strdup(tmp);
		tmp = strtok(NULL, delim);
	}

	free(content);
	return 0;
}

int yahoo_parsepacket_calendar(struct yahoo_context *ctx, struct yahoo_packet *pkt, struct yahoo_rawpacket *inpkt)
{
	char *content;
	char *tmp, delim[5];

	/* Make working copy of content */
	content = strdup(inpkt->content);

	/* init elements to all null */
	pkt->cal_url = NULL;
	pkt->cal_timestamp = NULL;
	pkt->cal_type = 0;
	pkt->cal_title = NULL;
	pkt->cal_description = NULL;

	tmp = NULL;
	delim[0] = 2;				/* control-b */
	delim[1] = 0;

	if (content)
	{
		tmp = strtok(content, delim);
	}

	if (tmp)					/* got the url */
	{
		pkt->cal_url = strdup(tmp);
		tmp = strtok(NULL, delim);

/*
   v= is not the type code
   i= doesn't look like it either
   tmp2 = strstr(pkt->cal_url, "v=");
   if ( tmp2 )
   {
   pkt->cal_type = atoi(tmp2);
   }
 */

	}

	if (tmp)					/* unknown (type code?) */
	{
/* appears this isn't it either, I don't see where it is */
/*      pkt->cal_type = atoi(tmp);  */
		tmp = strtok(NULL, "\r\n");
	}

	if (tmp)					/* timestamp */
	{
		pkt->cal_timestamp = strdup(tmp);
		tmp = strtok(NULL, "\r\n");
	}

	if (tmp)					/* title */
	{
		pkt->cal_title = strdup(tmp);
		tmp = strtok(NULL, delim);	/* use delim since it won't occur again */
	}

	if (tmp)
	{
		pkt->cal_description = strdup(tmp);
	}

	free(content);
	return 0;
}

int yahoo_parsepacket_chatinvite(struct yahoo_context *ctx, struct yahoo_packet *pkt, struct yahoo_rawpacket *inpkt)
{
	char *content;
	int len;

	/* Make working copy of content */
	content = strdup(inpkt->content);
	len = strlen(content);

	/* do special parsing for invite later on */
	pkt->chat_invite_content = strdup(content);

	return 0;
}

int yahoo_parsepacket_newcontact(struct yahoo_context *ctx, struct yahoo_packet *pkt, struct yahoo_rawpacket *inpkt)
{
	char *content;
	int len;

	/* Make working copy of content */
	content = strdup(inpkt->content);
	len = strlen(content);

	/* cheat for now, say if first digit is number */
	if (len > 0)
	{
		if (isdigit(content[0]))
		{
			return yahoo_parsepacket_status(ctx, pkt, inpkt);
		}
		else
		{
			return yahoo_parsepacket_message(ctx, pkt, inpkt);
		}
	}

	return 0;
}

int yahoo_parsepacket_status(struct yahoo_context *ctx, struct yahoo_packet *pkt, struct yahoo_rawpacket *inpkt)
{
	char *content;
	char *tmpc;
	char *tmpc2;
	char *tmp1;
	int i;
	int len;
	int index;

	/* Make working copy of content */
	content = strdup(inpkt->content);
	len = strlen(content);

// printf("raw content = '%s'\n", content);

	/* Pull off the flag from the initial part of the content */
	pkt->flag = 0;
	tmpc = content;
	while (tmpc[0] && isdigit(tmpc[0]))
	{
		pkt->flag = pkt->flag * 10 + (content[0] - '0');
		tmpc++;
	}
	if (tmpc[0] && tmpc[0] == ',')
	{
		tmpc++;
	}

	/* First, count the '(' characters to size the idstatus array */
	/* then allocate the array */
	pkt->idstatus_count = 0;
	for (i = 0; i < strlen(tmpc); i++)
	{
		if (tmpc[i] == '(')
		{
			pkt->idstatus_count++;
		}
	}

	/* print an error if I get the was not AWAY */
	if (strstr(tmpc, "was not AWAY"))
	{
		printf("Error: got a 'was not AWAY' message\n");
	}

	/* check for 'was not AWAY' in case calling app didn't obey
	   protocol for away/back setting */
	if (pkt->idstatus_count == 0 || strstr(tmpc, "was not AWAY"))
	{
		/* No entries, so no array needed */
		pkt->idstatus = NULL;
	}
	else
	{
		/* Allocate the array */
		pkt->idstatus = (struct yahoo_idstatus **)
			calloc(sizeof(struct yahoo_idstatus), pkt->idstatus_count);

		for (i = 0; i < pkt->idstatus_count; i++)
		{
			pkt->idstatus[i] = (struct yahoo_idstatus *)
				malloc(sizeof(struct yahoo_idstatus));
		}
	}

	index = 0;
	while (tmpc[0] && pkt->idstatus)
	{
		struct yahoo_idstatus *tmpid;

		/* Get pointer to allocated structure to hold status data */
		tmpid = pkt->idstatus[index++];

		/* Get pointer to start of next entry */
		tmpc2 = tmpc;
		while (tmpc2[0] && tmpc2[0] != ')')
		{
			tmpc2++;
		}
		if (tmpc2[0] == ')')
		{
			tmpc2++;
		}
		if (tmpc2[0] == ',')
		{
			tmpc2++;
		}

		/* nick(status,connection_id,unknown,in_pager,in_chat,directconn) */
		/* rewrite this whole section in a less ugly fashion */
		/* first pull off the id */
		tmp1 = strtok(tmpc, "(");
		if (tmp1)
		{
			tmpid->id = strdup(tmp1);
			for (i = 0; i < 6 && tmp1; i++)
			{
				tmp1 = strtok(NULL, ",");

				/* then pull off the particular element of the list */
				if (tmp1)
				{
					switch (i)
					{
						case 0:	/* status */
							tmpid->status = atoi(tmp1);
							break;
						case 1:	/* session id */
							tmpid->connection_id = strdup(tmp1);
							break;
						case 2:	/* ip addr */
							break;
						case 3:	/* logged on? */
							tmpid->in_pager = atoi(tmp1);
							break;
						case 4:	/* chat log on */
							tmpid->in_chat = atoi(tmp1);
							break;
						case 5:	/* direct connection support? */
							tmpid->directconn = atoi(tmp1);
							break;
					}
				}
			}
		}

/* insert stuff to pull out chat logon status */

		/* move to beginning of next entry */
		tmpc = tmpc2;
	}

	/* Free working copy of content */
	free(content);

	/* Return ok for success */
	return (0);
}

int yahoo_parsepacket_message(struct yahoo_context *ctx, struct yahoo_packet *pkt, struct yahoo_rawpacket *inpkt)
{
	char *content;
	char *tmp_id;
	char *tmp_flag;
	int i, j, section;

	/* Make working copy of content */
	content = strdup(inpkt->content);
	tmp_id = strdup(content);
	tmp_flag = strdup(content);

	/* initialize */
	pkt->msg_status = 0;

	/* possible message content formats: */
	/*     userid(#) */
	/*     userid,,msg */

	i = 0;
	j = 0;
	section = 0;
	tmp_id[0] = 0;
	tmp_flag[0] = 0;
	while (i < strlen(content))
	{
		char ch = content[i];

		if (section == 0)		/* parsing userid */
		{
			if (ch == ',')
			{
				j = 0;
				section = 1;
			}
			else if (ch == '(')
			{
				j = 0;
				section = 2;
			}
			else
			{
				tmp_id[j++] = ch;
				tmp_id[j] = 0;
			}
		}
		else if (section == 1)	/* parsing flag */
		{
			if (ch == ',')
			{
				j = 0;
				section = 3;
			}
			else
			{
				tmp_flag[j++] = ch;
				tmp_flag[j] = 0;
			}
		}
		else if (section == 2)	/* parsing status */
		{
			if (ch == ')')
			{
				j = 0;
				section = 3;
			}
			else
			{
				if (isdigit(ch))
				{
					pkt->msg_status *= 10;
					pkt->msg_status += ch - '0';
				}
			}
		}
		else
		{
			pkt->msg = strdup(&content[i]);
			break;
		}

		i++;
	}

	/* do stuff with extracted parts */
	pkt->msg_id = strdup(tmp_id);
	pkt->msg_flag = strdup(tmp_flag);

	/* free working variables */
	free(tmp_flag);
	free(tmp_id);
	free(content);

	/* Return ok for success */
	return (0);
}

int yahoo_getdata(struct yahoo_context *ctx)
{
	char buf[1000];
	int res;

	/* This is a http mode connection, so just send a ping to get any
	   new data from the server. */
	if (ctx->connect_mode == YAHOO_CONNECT_HTTP ||
		ctx->connect_mode == YAHOO_CONNECT_HTTPPROXY)
	{
		yahoo_sendcmd(ctx, YAHOO_SERVICE_PING, ctx->user, "");
		return (1);
	}

	/* this assumes that data is ready */
	/* Read from the connection to the server and get any new data */
	res = read(ctx->sockfd, buf, 1000);
	if (res == -1)
	{
		printf("YahooLib: Error reading data from server.\n");
		exit(1);
	}
	if (res > 0)
	{
		yahoo_addtobuffer(ctx, buf, res);
	}

	return (1);
}

void yahoo_addtobuffer(struct yahoo_context *ctx,
	char *data, int datalen)
{
	if (ctx->debug > 1)
	{
		yahoo_hexdump("yahoo_addtobuffer", data, datalen);
	}

	/* Check buffer, increase size if necessary */
	if (!ctx->io_buf || ((ctx->io_buf_maxlen - ctx->io_buf_curlen) < 500))
	{
		char *new_io_buf;

		ctx->io_buf_maxlen += 10240;
		new_io_buf = (char *) malloc(ctx->io_buf_maxlen);

		if (ctx->io_buf)
		{
			memcpy(new_io_buf, ctx->io_buf, ctx->io_buf_curlen);
			free(ctx->io_buf);
		}

		ctx->io_buf = new_io_buf;
	}

	memcpy(ctx->io_buf + ctx->io_buf_curlen, data, datalen);
	ctx->io_buf_curlen += datalen;
}

struct yahoo_rawpacket *yahoo_getpacket(struct yahoo_context *ctx)
{
	struct yahoo_rawpacket *pkt;
	struct yahoo_rawpacket *retpkt;
	int *buflen = &ctx->io_buf_curlen;
	char *buffer = ctx->io_buf;
	unsigned int contentlen;

	/* If buffer doesn't start with YHOO, skip bytes until it
	   does. This is to protect against possible packet alignment
	   errors if I size something wrong at any time. */

	while ((*buflen >= 4) && (memcmp(buffer, "YHOO", 4)))
	{
/* making quiet for now so I don't have to work too hard on the HTTP support */
#if 0
		printf("\nskipped buffer byte (%d)\n", buffer[0]);
#endif
		memmove(buffer, buffer + 1, *buflen - 1);
		*buflen = *buflen - 1;
	}

	/* Don't do anything if the buffer doesn't have at least a full 
	   header */
	if (*buflen < YAHOO_PACKET_HEADER_SIZE)
	{
// printf("returning null cause buffer is too small\n");
		return NULL;
	}

/* print out the beginning of the buffer */
#if 0
	printf("Buffer (buflen = %d):\n", *buflen);
	for (i = 0; i < *buflen; i++)
	{
		if ((i) % 10 == 0)
		{
			printf("\n%.4d: ", i);
		}
		if (isprint(buffer[i]))
		{
			printf("%-3d %c  ", buffer[i], buffer[i]);
		}
		else
		{
			printf("%-3d    ", buffer[i]);
		}
	}
	printf("\n");
#endif
	/* Make pkt point to buffer for ease of use */
	pkt = (struct yahoo_rawpacket *) buffer;

	/* Determine the content size specified by the header */
	contentlen = yahoo_makeint(pkt->len) - YAHOO_PACKET_HEADER_SIZE;
// printf("contentlen = %d\n", contentlen);

	/* Don't continue if buffer doesn't have full content in it */
	if (*buflen < (YAHOO_PACKET_HEADER_SIZE + contentlen))
	{
// printf("buffer not big enough for contentlen\n");
		return NULL;
	}

	/* Copy this packet */
	retpkt = (struct yahoo_rawpacket *) malloc(YAHOO_PACKET_HEADER_SIZE + contentlen);
	memcpy(retpkt, buffer, YAHOO_PACKET_HEADER_SIZE + contentlen);

	/* Shift the buffer */
	memmove(buffer, buffer + YAHOO_PACKET_HEADER_SIZE + contentlen,
		*buflen - YAHOO_PACKET_HEADER_SIZE - contentlen);

	/* Adjust the buffer length */
	*buflen -= (YAHOO_PACKET_HEADER_SIZE + contentlen);

	/* Return the packet */
	return retpkt;
}

#if 0
void cmd_addid()
{
	char user[36];

	printf("Add ID: ");
	fgets(user, 36, stdin);
	user[strlen(user) - 1] = 0;

	if (strlen(user))
		xmt_packet(servfd, YAHOO_SERVICE_ADDIDENT, username, user, user);
}

void cmd_ignore()
{
	char user[36];

	printf("Ignore: ");
	fgets(user, 36, stdin);
	user[strlen(user) - 1] = 0;

	if (strlen(user))
		xmt_packet(servfd, YAHOO_SERVICE_ADDIGNORE, username, username, user);
}

void cmd_whois()
{
	char user[36];

	printf("Whois: ");
	fgets(user, 36, stdin);
	user[strlen(user) - 1] = 0;

	if (strlen(user))
		xmt_packet(servfd, YAHOO_SERVICE_FETCHSTATE, username, username, user);
}

#endif

int yahoo_tcp_readline(char *ptr, int maxlen, int fd)
{
	int n, rc;
	char c;

	for (n = 1; n < maxlen; n++)
	{
	  again:

		if ((rc = read(fd, &c, 1)) == 1)
		{
			*ptr++ = c;
			if (c == '\n')
				break;
		}
		else if (rc == 0)
		{
			if (n == 1)
				return (0);		/* EOF, no data */
			else
				break;			/* EOF, w/ data */
		}
		else
		{
			if (errno == EINTR)
				goto again;
			printf("Yahoo: Error reading from socket in yahoo_tcp_readline.\n");
			exit(1);
		}
	}

	*ptr = 0;
	return (n);
}

/* really ugly brute force approach - someone find a GPL'd/free
   equivalent and replace this p.o.s. */
char *yahoo_urlencode(char *data)
{
	static char *tmp = NULL;
	char buf[4];
	int i;

	if (tmp)
	{
		free(tmp);
		tmp = NULL;
	}
	if (!data)
	{
		return NULL;
	}

	tmp = (char *) malloc(3 * strlen(data) + 1);
	tmp[0] = 0;

	for (i = 0; i < strlen(data); i++)
	{
		sprintf(buf, "%%%.2X", data[i]);
		strcat(tmp, buf);
	}

	return tmp;
}

int yahoo_socket_connect(struct yahoo_context *ctx,
	char *host, int port)
{
	struct sockaddr_in serv_addr;
	struct hostent *server;
	int servfd;
	int res;

	if (!ctx || !host || !port)
	{
		return 0;
	}

	server = gethostbyname(host);
	if (!server)
	{
		printf("YahooLib: failed to look up server (%s:%d)\n", host, port);
		return (0);
	}

	servfd = socket(AF_INET, SOCK_STREAM, 0);

	bzero(&serv_addr, sizeof(serv_addr));
	serv_addr.sin_family = AF_INET;
	bcopy(server->h_addr, &serv_addr.sin_addr.s_addr, server->h_length);
	serv_addr.sin_port = htons(port);

	res = -1;
	if (ctx->connect_mode == YAHOO_CONNECT_SOCKS4)
	{
#if defined(WITH_SOCKS4)
		res = Rconnect(servfd, (struct sockaddr *) &serv_addr, sizeof(serv_addr));
#endif
	}
	else if (ctx->connect_mode == YAHOO_CONNECT_SOCKS5)
	{
#if defined(WITH_SOCKS5)
#endif
	}
	else if (ctx->connect_mode == YAHOO_CONNECT_NORMAL ||
			ctx->connect_mode == YAHOO_CONNECT_HTTP ||
		ctx->connect_mode == YAHOO_CONNECT_HTTPPROXY)
	{
		res = connect(servfd, (struct sockaddr *) &serv_addr, sizeof(serv_addr));
	}
	else
	{
		printf("YahooLib: unhandled connect mode (%d).\n", ctx->connect_mode);
		return (0);
	}

	if (res < 0)
	{
		close(servfd);
		printf("YahooLib: failed to connect to server (%s:%d)\n",
			host, port);
		return (0);
	}

	return servfd;
}

void yahoo_hexdump(char *label, unsigned char *data, int datalen)
{
	int i, j;
	int val, skipped_last;
	char current[100];
	char last[100];
	char tmp[15];
	char outline[100];
	static int last_datalen = 0;
	static unsigned char *last_data = NULL;

	if ( last_data )
	{
		if ( last_datalen == datalen && !memcmp(data,last_data,datalen))
		{
			printf("\n%s: <same as last dump>\n", label);
			return;
		}
		free(last_data);
	}

	/* Copy the packet so we can don't duplicate it next time. */
	last_datalen = datalen;
	last_data = malloc(datalen);
	memcpy(last_data,data,datalen);

	/* Handle printing the full entry out */
	printf("\n");
	printf("%s:\n", label);

	skipped_last = 0;
	last[0] = 0;
	for (j = 0; j * 16 < datalen; j++)
	{
		current[0] = 0;

		/* Print out in hex */
		for (i = j * 16; i < (j * 16 + 16); i++)
		{
			if (i < datalen)
			{
				val = data[i];
				sprintf(tmp, "%.2X ", val);
			}
			else
			{
				sprintf(tmp, "   ");
			}
			strcat(current, tmp);
		}

		/* Print out in ascii */
		strcat(current, "  ");
		for (i = j * 16; i < (j * 16) + 16; i++)
		{
			if (i < datalen)
			{
				if (isprint(data[i]))
				{
					sprintf(tmp, "%c", data[i]);
				}
				else
				{
					sprintf(tmp, ".");
				}
			}
			else
			{
				sprintf(tmp, " ");
			}
			strcat(current, tmp);
		}

		outline[0] = 0;
		if (!strcmp(current, last))
		{
			if (!skipped_last)
			{
				strcpy(outline, " ....:\n");
			}
			skipped_last = 1;
		}
		else
		{
			sprintf(outline, " %.4d:  %s\n", j * 16, current);
			skipped_last = 0;
		}
		printf("%s", outline);
		strcpy(last, current);
	}

	if (skipped_last)
	{
		printf("%s", outline);
	}
	printf("\n");

}
